# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: event.py
#  Purpose: Expansion of the obspy.core.event module
#   Author: uquake development team
#    Email: devs@uquake.org
#
# Copyright (C) 2016 uquake development team
# --------------------------------------------------------------------
"""
Expansion of the obspy.core.event module

:copyright:
    uquake development team (devs@uquake.org)
:license:
    GNU Lesser General Public License, Version 3
    (http://www.gnu.org/copyleft/lesser.html)
"""

import base64
import io
import warnings

import numpy as np
import obspy.core.event as obsevent
from obspy.core.event import *
from obspy.core.util import AttribDict
from copy import deepcopy
from base64 import b64encode, b64decode
from io import BytesIO
import pickle
from uquake.waveform.mag_utils import calc_static_stress_drop
from .logging import logger
from pathlib import Path

debug = False


class Catalog(obsevent.Catalog):
    extra_keys = []

    __doc__ = obsevent.Catalog.__doc__.replace('obspy', 'uquake')

    def __init__(self, obspy_obj=None, **kwargs):
        # _init_handler(self, obspy_obj, **kwargs)
        if obspy_obj and len(kwargs) > 0:
            raise AttributeError("Initialize from either \
                                  obspy_obj or kwargs, not both")

        if obspy_obj:

            for key in obspy_obj.__dict__.keys():
                if key == 'events':
                    events = []
                    for event in obspy_obj:
                        events.append(Event(obspy_obj=event))
                    self.events = events
                else:
                    self.__dict__[key] = obspy_obj.__dict__[key]

        else:
            super(type(self), self).__init__(
                **kwargs)

    def __setattr__(self, name, value):
        super(type(self), self).__setattr__(name, value)

    def write(self, fileobj, format='quakeml', **kwargs):
        for event in self.events:
            for ori in event.origins:
                for ar in ori.arrivals:
                    if 'extra' in ar.keys():
                        del ar.extra

        result = obsevent.Catalog.write(self, fileobj, format=format, **kwargs)

        for event in self.events:
            for ori in event.origins:
                ars = []
                for ar in ori.arrivals:
                    ars.append(Arrival(obspy_obj=ar))
                ori.arrivals = ars

        return result

    def duplicate(self):
        """
        this function duplicates a catalog object, this function does not
        duplicate picks. It creates an object containing multiple event
        containing each one origin and one magnitude.
        :return: a new catalog object
        """

        new_events = []
        for event in self.events:
            if len(event.origins) == 0:
                new_origins = []
                preferred_origin_id = None

            else:
                new_origins = [event.preferred_origin() or event.origins[-1]]

                preferred_origin_id = ResourceIdentifier()
                new_origins[0].resource_id = preferred_origin_id
                new_origins[0].arrivals = []
                new_rays = []
                for ray in new_origins[0].rays:
                    ray.resource_id = ResourceIdentifier()
                    new_rays.append(ray)
                new_origins[0].rays = new_rays

            if len(event.magnitudes) == 0:
                new_magnitudes = []
                preferred_magnitude_id = None
            else:
                new_magnitudes = [event.preferred_magnitude()
                                  or event.magnitudes[-1]]

                preferred_magnitude_id = ResourceIdentifier()
                new_magnitudes[0].resource_id = preferred_magnitude_id
                new_magnitudes[0].origin_id = preferred_origin_id

            new_event = Event(origins=new_origins, magnitudes=new_magnitudes)
            new_event.preferred_origin_id = preferred_origin_id
            new_event.preferred_magnitude_id = preferred_magnitude_id

            new_events.append(new_event)

            return Catalog(events=new_events)

    def copy(self):
        return deepcopy(self)


class Event(obsevent.Event):
    # _format keyword is actualy a missing obspy default
    extra_keys = ['_format', 'ACCEPTED', 'ASSOC_SEISMOGRAM_NAMES',
                  'AUTO_PROCESSED',
                  'BLAST', 'CORNER_FREQUENCY', 'DYNAMIC_STRESS_DROP',
                  'ENERGY', 'ENERGY_P', 'ENERGY_S', 'EVENT_MODIFICATION_TIME',
                  'EVENT_NAME', 'EVENT_TIME_FORMATTED', 'EVENT_TIME_NANOS',
                  'LOCAL_MAGNITUDE', 'LOCATION_RESIDUAL', 'LOCATION_X',
                  'LOCATION_Y', 'LOCATION_Z', 'MANUALLY_PROCESSED',
                  'NUM_ACCEPTED_TRIGGERS', 'NUM_TRIGGERS', 'POTENCY',
                  'POTENCY_P', 'POTENCY_S', 'STATIC_STRESS_DROP', 'TAP_TEST',
                  'TEST', 'TRIGGERED_SITES', 'USER_NAME', 'network']

    __doc__ = obsevent.Event.__doc__.replace('obspy', 'uquake')

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)

    def __setattr__(self, name, value):
        _set_attr_handler(self, name, value)

    def __str__(self):
        return "Event:\t%s\n\n%s" % (
            self.short_str(),
            "\n".join(super(Event, self).__str__().split("\n")[1:]))

    def short_str(self):
        out = ''

        if self.origins:
            og = self.preferred_origin() or self.origins[-1]

            if og.x is None:
                x = 0
            else:
                x = og.x
            if og.y is None:
                y = 0
            else:
                y = og.y
            if og.z is None:
                z = 0
            else:
                z = og.z

            out += '%s | %0.2f, %0.2f, %0.2f | %s' % (og.time,
                                                      x, y, z,
                                                      og.evaluation_mode)

        if self.magnitudes:
            magnitude = self.preferred_magnitude() or self.magnitudes[-1]
            out += ' | %0.2f %-2s' % (magnitude.mag,
                                      magnitude.magnitude_type)

        return out

        self.picks += picks

    def write(self, fileobj, **kwargs):
        for ori in self.origins:
            arrivals = []
            for ar in ori.arrivals:
                if 'extra' in ar.keys():
                    del ar.extra

        return obsevent.Event.write(self, fileobj, **kwargs)

    def plot_focal_mechanism(self):
        pass

    def append_origin_as_preferred_origin(self, new_origin: Origin):
        self.origins.append(new_origin)
        self.preferred_origin_id = new_origin.resource_id


class Origin(obsevent.Origin):
    __doc__ = obsevent.Origin.__doc__.replace('obspy', 'uquake')
    extra_keys = ['x', 'y', 'z', 'x_error', 'y_error', 'z_error', 'scatter',
                  '__encoded_rays__', 'author']

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)

    def __setattr__(self, name, value):

        if name == 'rays':
            self.__encoded_rays__ = self.__encode_rays__(self, value)
        # elif name == 'encoded_rays':
        #     self.__encoded_rays__ = value
        else:
            _set_attr_handler(self, name, value)

    def __getattr__(self, item):
        if item == 'rays':
            try:
                return self.__decode_rays__(self)
            except pickle.UnpicklingError:
                self.__encoded_rays__ = eval(self.__encoded_rays__)
                return self.__decode_rays__(self)

        else:
            return self.__dict__[item]

    @staticmethod
    def __encode_rays__(self, rays):
        return b64encode(pickle.dumps(rays))

    @staticmethod
    def __decode_rays__(self):
        if self.__encoded_rays__ is None:
            return
        try:
            return pickle.loads(b64decode(self.__encoded_rays__))
        except Exception as e:
            self.__encoded_rays__ = eval(self.__encoded_rays__)
            return pickle.loads(b64decode(self.__encoded_rays__))

    def get_arrival_id(self, phase, station_code):
        arrival_id = None
        for arrival in self.arrivals:
            if (arrival.phase == phase) and (arrival.get_sta() ==
                                             station_code):
                arrival_id = arrival.resource_id

        return arrival_id

    def append_ray(self, item):
        if self.rays is None:
            self.rays = [item]
        else:
            self.rays = self.rays + [item]

    @property
    def rms_residual(self):
        if len(self.arrivals) == 0:
            return None
        residuals = [arrival.time_residual ** 2 for arrival in self.arrivals]
        return np.sqrt(np.mean(residuals))

    @property
    def loc(self):
        return np.array([self.x, self.y, self.z])

    @property
    def location_uncertainty(self):
        if self.origin_uncertainty is None:
            return None
        else:
            if self.origin_uncertainty.confidence_ellipsoid is None:
                return None
            return self.origin_uncertainty.confidence_ellipsoid \
                .semi_major_axis_length

    def get_origin(self):
        if self.preferred_origin_id is not None:
            return self.preferred_origin_id.get_referred_object()

    def get_all_magnitudes_for_origin(self, cat):
        magnitudes = []

        for event in cat:
            for mag in event.magnitudes:
                if mag.origin_id.id == self.resource_id.id:
                    magnitudes.append(mag)

        return magnitudes

    def __str__(self, **kwargs):

        string = f"""
       resource_id: {self.resource_id}
         time(UTC): {self.time}
                 x: {self.x}
                 y: {self.y}
                 z: {self.z}
       uncertainty: {self.location_uncertainty} (m)
   evaluation_mode: {self.evaluation_mode}
 evaluation_status: {self.evaluation_status}
                ---------
          arrivals: {len(self.arrivals):d}
        """
        return string

    def get_incidence_baz_angles(self, site_code, phase):
        baz = None
        inc = None
        for ray in self.rays:
            if (ray.site_code == site_code) and (ray.phase == phase):
                baz = ray.back_azimuth
                inc = ray.incidence_angle
                break
        return baz, inc

    def get_ray_station_phase(self, site_code, phase):
        out_ray = None
        for ray in self.rays:
            if (ray.site_code == site_code) and (ray.phase == phase):
                out_ray = ray
                break
        return out_ray

    def distance_station(self, site_code, phase='P'):
        ray = self.get_ray_station_phase(self, site_code, phase)
        if ray is None:
            return None

        return ray.length


class Magnitude(obsevent.Magnitude):
    __doc__ = obsevent.Magnitude.__doc__.replace('obspy', 'uquake')

    extra_keys = ['energy_joule', 'energy_p_joule', 'energy_p_std',
                  'energy_s_joule', 'energy_s_std', 'corner_frequency_hz',
                  'corner_frequency_p_hz', 'corner_frequency_s_hz',
                  'corner_frequency_error',
                  'time_domain_moment_magnitude',
                  'frequency_domain_moment_magnitude',
                  'moment_magnitude', 'moment_magnitude_uncertainty',
                  'seismic_moment', 'potency_m3', 'source_volume_m3',
                  'apparent_stress', 'static_stress_drop_mpa',
                  'quick_magnitude', 'error']

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)

    def __setattr__(self, name, value):
        _set_attr_handler(self, name, value)

    @classmethod
    def from_dict(cls, input_dict, origin_id=None, obspy_obj=None, **kwargs):
        out_cls = cls(obspy_obj=obspy_obj, **input_dict,
                      origin_id=origin_id, **kwargs)
        out_cls.mag = input_dict['moment_magnitude']
        out_cls.magnitude_type = 'Mw'

        return out_cls

    @property
    def static_stress_drop_mpa(self):
        ssd = None
        if self.magnitude_type == "Mw":
            if self.mag and self.corner_frequency_hz:
                ssd = calc_static_stress_drop(self.mag,
                                              self.corner_frequency_hz)
        return ssd

    @property
    def apparent_stress(self):
        app_stress = None
        if self.magnitude_type == 'Mw':
            if self.energy_joule and self.mag:
                app_stress = 2 * self.energy_joule / self.potency_m3
        return app_stress

    @property
    def seismic_moment(self):
        seismic_moment = None
        if self.magnitude_type == 'Mw':
            seismic_moment = 10 ** (3 * (self.mag + 6.02) / 2)
        return seismic_moment

    # @seismic_moment.setter
    # def seismic_moment(self, val):
    #     self.mag = val
    #     self.magnitude_type = 'Mw'

    @property
    def potency_m3(self):
        potency = None
        mu = 29.5e9
        if self.magnitude_type == 'Mw':
            if self.mag:
                potency = self.seismic_moment / mu
        return potency

    def __str__(self, **kwargs):

        es_ep = None
        if self.energy_p_joule and self.energy_s_joule:
            es_ep = self.energy_s_joule / self.energy_p_joule

        string = f"""
             resource_id: {self.resource_id.id}     
               Magnitude: {self.mag}
          Magnitude type: {self.magnitude_type}
   Corner frequency (Hz): {self.corner_frequency_hz}
 Radiated Energy (joule): {self.energy_joule}
                   Es/Ep: {es_ep}
          Seismic moment: {self.seismic_moment}
       Source volume(m3): {self.potency_m3}
Static stress drop (MPa): {self.static_stress_drop_mpa}
     Apparent stress(Pa): {self.apparent_stress}
         evaluation mode: {self.evaluation_mode}
       evaluation status: {self.evaluation_status}
        """

        return string


class Pick(obsevent.Pick):
    __doc__ = obsevent.Pick.__doc__.replace('obspy', 'uquake')
    extra_keys = ['method', 'snr', 'trace_id', 'author', 'linearity',
                  'planarity', 'azimuth', 'dip']

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)
        if obspy_obj:
            wid = self.waveform_id
            self.trace_id = "%s.%s.%s.%s" % (wid.network_code,
                                             wid.station_code,
                                             wid.location_code,
                                             wid.channel_code)

    def __setattr__(self, name, value):
        _set_attr_handler(self, name, value)

    def __str__(self, **kwargs):
        string = """
          trace_id: %s
              time: %s
             phase:[%s]
            method: %s [%s]
   evaluation_mode: %s
 evaluation_status: %s
       resource_id: %s
        """ \
                 % (self.trace_id, self.time.strftime("%Y/%m/%d %H:%M:%S.%f"),
                    self.phase_hint, self.method, self.snr,
                    self.evaluation_mode,
                    self.evaluation_status, self.resource_id)
        return string

    def get_sta(self):
        if self.waveform_id is not None:
            return self.waveform_id.station_code

    @property
    def site(self):
        if self.waveform_id is not None:
            site = self.waveform_id.station_code + \
                   self.waveform_id.location_code
            return site

    @property
    def site_code(self):
        return self.site


class WaveformStreamID(obsevent.WaveformStreamID):
    __doc__ = obsevent.WaveformStreamID.__doc__.replace('obspy', 'uquake')
    extra_keys = []

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)

    @property
    def site_code(self):
        site = self.station_code + self.location_code
        return site

    @property
    def site(self):
        return self.site_code


class Arrival(obsevent.Arrival):
    __doc__ = obsevent.Arrival.__doc__.replace('obspy', 'uquake')

    extra_keys = ['ray', 'backazimuth', 'inc_angle',
                  'peak_vel', 'tpeak_vel', 't1', 't2', 'pulse_snr',
                  'peak_dis', 'tpeak_dis', 'max_dis', 'tmax_dis',
                  'dis_pulse_width', 'dis_pulse_area', 'smom', 'fit',
                  'tstar', 'hypo_dist_in_m', 'vel_flux', 'vel_flux_Q',
                  'energy', 'fmin', 'fmax', 'traces']

    def __init__(self, obspy_obj=None, **kwargs):
        _init_handler(self, obspy_obj, **kwargs)

    def __setattr__(self, name, value):
        if name == 'polarity':
            if value in ['positive', 'negative', 'undecidable']:
                self.get_pick().polarity = value
            elif value == -1:
                self.get_pick().polarity = 'negative'
            elif value == 1:
                self.get_pick().polarity = 'positive'
            else:
                self.get_pick().polarity = 'undecidable'

        else:
            _set_attr_handler(self, name, value)

    def __repr__(self):
        out_str = f"""
       resource_id: {self.resource_id}
           pick_id: {self.get_pick().resource_id}
             phase: {self.phase}
           azimuth: {self.azimuth} (deg)
          distance: {self.distance} (m)
     takeoff_angle: {self.takeoff_angle} (deg)
     time_residual: {self.time_residual * 1000} (ms)
       time_weight: {self.time_weight}
        """
        return out_str

    @staticmethod
    def calculate_time_residual(observed: float, predicted: float):
        return observed - predicted

    @property
    def polarity(self):
        if self.get_pick().polarity == 'positive':
            return 1.0
        elif self.get_pick().polarity == 'negative':
            return -1.0
        else:
            return None

    @property
    def pick(self):
        return self.get_pick()

    @property
    def site_code(self):
        return self.pick.site

    @property
    def site(self):
        return self.pick.site

    def get_pick(self):
        if self.pick_id is not None:
            return self.pick_id.get_referred_object()

    def get_sta(self):
        if self.pick_id is not None:
            pick = self.pick_id.get_referred_object()
            return pick.get_sta()


def get_arrival_from_pick(arrivals, pick):
    """
      return arrival corresponding to pick

      :param arrivals: list of arrivals
      :type arrivals: list of either obspy.core.event.origin.Arrival
                      or uquake.core.event.origin.Arrival
      :param pick: P or S pick
      :type pick: either obspy.core.event.origin.Pick
                      or uquake.core.event.origin.Pick
      :return arrival
      :rtype: obspy.core.event.origin.Arrival or
              uquake.core.event.origin.Arrival
    """

    arrival = None
    for arr in arrivals:
        if arr.pick_id == pick.resource_id:
            arrival = arr
            break

    return arrival


def read_events(filename, **kwargs):
    if isinstance(filename, Path):
        filename = str(filename)
    # converting the obspy object into uquake objects

    cat = obsevent.read_events(filename, **kwargs)
    mq_catalog = Catalog(obspy_obj=cat)

    return mq_catalog


def _init_handler(self, obspy_obj, **kwargs):
    """
    Handler to initialize uquake objects which
    inherit from ObsPy class. If obspy_obj is none,
    Kwargs is expected to be a mix of obspy kwargs
    and uquake kwargs specified by the hardcoded
    extra_keys.
    """

    if obspy_obj and len(kwargs) > 0:
        raise AttributeError("Initialize from either \
                              obspy_obj or kwargs, not both")

    # default initialize the extra_keys args to None
    self['extra'] = {}
    [self.__setattr__(key, None) for key in self.extra_keys]

    if obspy_obj:
        _init_from_obspy_object(self, obspy_obj)

        if 'resource_id' in obspy_obj.__dict__.keys():
            rid = obspy_obj.resource_id.id
            self.resource_id = ResourceIdentifier(id=rid,
                                                  referred_object=self)
    else:
        extra_kwargs = pop_keys_matching(kwargs, self.extra_keys)
        super(type(self), self).__init__(**kwargs)  # init obspy_origin args
        [self.__setattr__(k, v) for k, v in extra_kwargs.items()]  # init
        # extra_args


def _init_from_obspy_object(uquake_obj, obspy_obj):
    """
    When initializing uquake object from obspy_obj
    checks attributes for lists of obspy objects and
    converts them to equivalent uquake objects.
    """

    class_equiv = {obsevent.event: Event,
                   obsevent.Pick: Pick,
                   obsevent.Arrival: Arrival,
                   obsevent.Origin: Origin,
                   obsevent.Magnitude: Magnitude,
                   obsevent.WaveformStreamID: WaveformStreamID}

    for key, val in obspy_obj.__dict__.items():
        itype = type(val)
        if itype in class_equiv:
            uquake_obj.__setattr__(key, class_equiv[itype](val))

        elif itype == list:
            out = []
            for item in val:
                itype = type(item)
                if itype in class_equiv:
                    out.append(class_equiv[itype](item))
                else:
                    out.append(item)
            uquake_obj.__setattr__(key, out)
        else:
            uquake_obj.__setattr__(key, val)


def _set_attr_handler(self, name, value, namespace='UQUAKE'):
    """
    Generic handler to set attributes for uquake objects
    which inherit from ObsPy objects. If 'name' is not in
    default keys then it will be set in self['extra'] dict. If
    'name' is not in default keys but in the self.extra_keys
    then it will also be set as a class attribute. When loading
    extra keys from quakeml file, those in self.extra_keys will
    be set as attributes.
    """

    #  use obspy default setattr for default keys
    if name in self.defaults.keys():
        self.__dict__[name] = value
        if isinstance(self.__dict__[name], ResourceIdentifier):
            self.__dict__[name] = ResourceIdentifier(id=value.id)
        # super(type(self), self).__setattr__(name, value)
    elif name in self.extra_keys:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self[name] = value
        if type(value) is np.ndarray:
            value = "npy64_" + array_to_b64(value)
        elif type(value) is str:
            if "npy64_" in value:
                value.replace("npy64_", "")
                b64_to_array(value)
        self['extra'][name] = {'value': value, 'namespace': namespace}
    # recursive parse of 'extra' args when constructing uquake from obspy
    elif name == 'extra':
        if 'extra' not in self:  # hack for deepcopy to work
            self['extra'] = {}
        for key, adict in value.items():
            if key in self.extra_keys:
                self.__dict__[key] = parse_string_val(adict.value)
            else:
                self['extra'][key] = adict
    else:
        raise KeyError(name)


def _set_attr_handler2(self, name, value, namespace='UQUAKE'):
    """
    Generic handler to set attributes for uquake objects
    which inherit from ObsPy objects. If 'name' is not in
    default keys then it will be set in self['extra'] dict. If
    'name' is not in default keys but in the self.extra_keys
    then it will also be set as a class attribute. When loading
    extra keys from quakeml file, those in self.extra_keys will
    be set as attributes.
    """

    #  use obspy default setattr for default keys
    if name in self.defaults.keys():
        super(type(self), self).__setattr__(name, value)
    # recursive parse of extra args when constructing uquake from obspy
    elif name == 'extra':
        if 'extra' not in self:  # hack for deepcopy to work
            self['extra'] = {}
        for key, adict in value.items():
            self.__setattr__(key, parse_string_val(adict.value))
    else:  # branch for extra keys
        if name in self.extra_keys:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self[name] = value
        if type(value) is np.ndarray:
            value = "npy64_" + array_to_b64(value)
        self['extra'][name] = {'value': value, 'namespace': namespace}


def isfloat(value):
    try:
        float(value)
        return True
    except ValueError:
        return False


def pop_keys_matching(dict_in, keys):
    # Move keys from dict_in to dict_out
    dict_out = {}
    for key in keys:
        if key in dict_in:
            dict_out[key] = dict_in.pop(key)
    return dict_out


def array_to_b64(array):
    output = io.BytesIO()
    np.save(output, array)
    content = output.getvalue()
    encoded = base64.b64encode(content).decode('utf-8')
    return encoded


def b64_to_array(b64str):
    arr = np.load(io.BytesIO(base64.b64decode(b64str)))
    return arr


def parse_string_val(val, arr_flag='npy64_'):
    """
    Parse extra args in quakeML which are all stored as string.
    """
    if val is None:  # hack for deepcopy ignoring isfloat try-except
        val = None
    elif type(val) == AttribDict:
        val = val
    elif isfloat(val):
        val = float(val)
    elif str(val) == 'None':
        val = None
    elif val[:len(arr_flag)] == 'npy64_':
        val = b64_to_array(val[len(arr_flag):])
    return val


class RayCollection:
    def __init__(self, rays=[]):
        self.__encoded_rays__ = self.__encode_rays__(self, rays)
        self.origin_id = None

    def __setattr__(self, key, value):
        if key == 'rays':
            self.__encoded_rays__ = self.__encode_rays__(self, value)
        else:
            self.__dict__[key] = value

    def __getattr__(self, item):
        if item == 'rays':
            return self.__decode_rays__(self)
        else:
            return self.__dict__[item]

    @staticmethod
    def __encode_rays__(self, rays):
        return b64encode(pickle.dumps(rays))

    @staticmethod
    def __decode_rays__(self):
        return pickle.loads(b64decode(self.__encoded_rays__))

    def append(self, item):
        self.rays = self.rays + [item]


class Ray:

    def __init__(self, nodes: list = [], site_code: str = None,
                 arrival_id: ResourceIdentifier = None,
                 phase: str = None, azimuth: float = None,
                 takeoff_angle: float = None,
                 travel_time: float = None,
                 earth_model_id: ResourceIdentifier = None,
                 network: str = None, station: str = None,
                 location: str = None):
        """
        :param nodes: ray nodes
        :param site_code: site code
        :param arrival_id: the ResourceIdentifier of the arrival associated to
        the ray
        :param phase: seismic phase ("P" or "S")
        :param azimuth: Azimuth in degrees
        :param takeoff_angle: takeoff angle in degrees
        :param travel_time: travel time between the source and the site in
        second
        :param earth_model_id: velocity model ResourceIdentifier
        """

        self.nodes = np.array(nodes)
        self.site_code = site_code
        self.arrival_id = arrival_id
        self.phase = phase
        self.azimuth = azimuth
        self.takeoff_angle = takeoff_angle
        self.travel_time = travel_time
        self.resource_id = obsevent.ResourceIdentifier()
        self.earth_model_id = earth_model_id

    def __setattr__(self, key, value):
        if key == 'phase':
            if value is None:
                self.__dict__[key] = value
            else:
                self.__dict__[key] = value.upper()
        else:
            self.__dict__[key] = value

    @property
    def length(self):
        if len(self.nodes) < 2:
            return 0

        length = 0
        for k, node1 in enumerate(self.nodes[0:-1]):
            node2 = self.nodes[k + 1]
            length += np.linalg.norm(node1 - node2)

        return length

    @property
    def baz(self):
        # back_azimuth
        baz = None
        if len(self.nodes) > 2:
            v = self.nodes[-2] - self.nodes[-1]
            baz = np.arctan2(v[0], v[1]) * 180 / np.pi
        return baz

    @property
    def back_azimuth(self):
        self.baz

    @property
    def incidence_angle(self):
        ia = None
        if len(self.nodes) > 2:
            v = self.nodes[-2] - self.nodes[-1]
            h = np.sqrt(v[0] ** 2 + v[1] ** 2)
            ia = np.arctan2(h, v[2]) * 180 / np.pi
        return ia

    def __len__(self):
        return len(self.nodes)

    def __str__(self):
        txt = \
            f"""
       site code: {self.site_code}
        arrival id: {self.arrival_id}
             phase: {self.phase}
        length (m): {self.length}
   number of nodes: {len(self.nodes)}
            """
        return txt

    def __repr__(self):
        return self.__str__()


def break_down(event):
    origin = event.origins[0]
    print("break_down: Here's what obspy reads:")
    print(origin)
    print("origin res id: %s" % origin.resource_id.id)
    print("id(origin): %s" % id(origin))
    print("id(origin.resource_id):%s" % id(origin.resource_id))
    ref_obj = origin.resource_id.get_referred_object()
    print("id(ref_obj):%s" % id(ref_obj))

    return


# MTH: this could(should?) be moved to waveforms/pick.py ??
def make_pick(time, phase='P', wave_data=None, snr=None, mode='automatic',
              status='preliminary', method_string=None, resource_id=None):
    this_pick = Pick()
    this_pick.time = time
    this_pick.phase_hint = phase
    this_pick.evaluation_mode = mode
    this_pick.evaluation_status = status

    this_pick.method = method_string
    this_pick.snr = snr

    if wave_data is not None:
        this_pick.waveform_id = WaveformStreamID(
            network_code=wave_data.stats.network,
            station_code=wave_data.stats.station,
            location_code=wave_data.stats.location,
            channel_code=wave_data.stats.channel)

        this_pick.trace_id = wave_data.get_id()

    return this_pick



import numpy as np
import h5py


class H5TTable(object):
    """docstring for H5TTable"""

    def __init__(self, path, dset_key=None):
        self.path = path
        self.hf = h5py.File(path, 'r')
        self.keys = list(self.hf.keys())

        self.dset = None

        if dset_key is not None:
            self.set_dataset(dset_key)

        self.stations = self.hf['stations'][:].astype('U4')
        self._stadict = dict(zip(self.stations, np.arange(len(self.stations))))

        self.locations = self.hf['locations'][:]
        self.coords = self.hf['grid_locs'][:]

    def set_dataset(self, key):
        if key in self.keys:
            self.dset = self.hf[key]
        else:
            raise KeyError('dataset %s does not exist' % key)

    @property
    def shape(self):
        return self.hf.attrs['shape']

    @property
    def origin(self):
        return self.hf.attrs['origin']

    @property
    def spacing(self):
        return self.hf.attrs['spacing']

    def index_sta(self, stations):
        if isinstance(stations, (list, np.ndarray)):
            return np.array([self._stadict[sta] for sta in stations])
        else:
            return self._stadict[stations]

    def icol_to_xyz(self, index):
        nx, ny, nz = self.shape
        iz = index % nz
        iy = ((index - iz) // nz) % ny
        ix = index // (nz * ny)
        loc = np.array([ix, iy, iz], dtype=float) * self.spacing + self.origin

        return loc

    def xyz_to_icol(self, loc):
        x, y, z = loc
        ix, iy, iz = ((loc - self.origin) / self.spacing).astype(int)
        nx, ny, nz = self.shape
        # return (iz * nx * ny) + (iy * nx) + ix;

        return int((ix * ny * nz) + (iy * nz) + iz)

    def close(self):
        self.hf.close()


def gdef_to_points(shape, origin, spacing):
    maxes = origin + shape * spacing
    x = np.arange(origin[0], maxes[0], spacing).astype(np.float32)
    y = np.arange(origin[1], maxes[1], spacing).astype(np.float32)
    z = np.arange(origin[2], maxes[2], spacing).astype(np.float32)
    points = np.zeros((np.product(shape), 3), dtype=np.float32)
    # points = np.stack(np.meshgrid(x, y, z), 3).reshape(3, -1).astype(np.float32)
    ix = 0

    for xv in x:
        for yv in y:
            for zv in z:
                points[ix] = [xv, yv, zv]
                ix += 1

    return points



# -*- coding: utf-8; -*-
#
# (c) 2016 uquake development team
#
# This file is part of the uquake library
#
# uquake is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# uquake is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with uquake.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from ..core import Stream, Trace
from obspy.realtime.signal import kurtosis
from scipy.signal import detrend
from ..core.logging import logger
from ..core.event import make_pick, Pick
from ..core.inventory import Inventory
from ..core.util.tools import copy_picks_to_dict
from ..core import UTCDateTime


def measure_polarity(st, catalog, site, average_time_window=1e-3,
                     lp_filter_freq=100):
    """
    Measure the P- and S-wave polarities
    :param st: seismograms
    :type st: either obspy.core.Stream or uquake.core.Stream
    :param catalog: catalog object
    :type catalog: uquake.core.Catalog
    :param site: sensor information
    :type site: uquake.core.station.Site
    :param average_time_window: time window over which the polarity is measured (s)
    :type average_time_window: float
    :param lp_filter_freq: frequency of the low pass filter to apply for
    polarity picking
    :rparam: returns a copy of the input catalog
    :rtype: uquake.core.Stream
    """
    cat_out = catalog.copy()

    for iev, ev in enumerate(cat_out.events):
        for ipick, pick in enumerate(ev.picks):
            sta_code = pick.waveform_id.station_code
            ch_code = pick.waveform_id.channel_code
            trs = st.select(station=sta_code, channel=ch_code)
            trs = trs.detrend('linear').detrend('demean')
            sta = site.select(station=sta_code, channel=ch_code)
            if not sta.networks[0].stations:
                continue
            if (len(sta.networks) > 1) or (len(sta.networks[0].stations) > 1):
                logger.warning("MeasurePolarity: multiple station selected "
                               "from site. Only the first will be used")
            sta = sta.networks[0].stations[0]
            motion_type = sta.motion_type
            if motion_type == 'acceleration':
                trs = trs.integrate().detrend('linear').detrend(
                    'demean').integrate()
            elif motion_type == 'velocity':
                trs.integrate()
            elif motion_type == 'displacement':
                pass
            else:
                logger.warning("MeasurePolarity: motion_type not set for "
                               "sensor %s... Displacement will be assumed"
                               % stname)

            trs.filter('lowpass', freq=lp_filter_freq)
            if len(trs) > 1:
                logger.warning("number of trace for station %s and channel %s"
                               "is greater than 1. Only the first trace "
                               "will be used" % (
                                   sta_code, ch_code))
            tr = trs.traces[0]

            sp_s = int(
                (pick.time - tr.stats.starttime) * tr.stats.sampling_rate)
            sp_e = int(
                (sp_s + average_time_window * tr.stats.sampling_rate)) + 1
            pol = np.sign(np.mean(tr.data[sp_s:sp_e]) - tr.data[sp_s])

            if pol > 0:
                cat_out.events[iev].picks[ipick].polarity = "positive"
            elif pol < 0:
                cat_out.events[iev].picks[ipick].polarity = "negative"
            else:
                cat_out.events[iev].picks[ipick].polarity = "undecidable"

    return cat_out


def _measure_pick_polarity_tr(tr, pick_time, signal_type='Acceleration',
                              nsamp_avg=10):
    """Mesure pick polarity
    The pick polarity is measured on the displacement trace looking at the
    difference
    between the amplitude at the pick time and the sign of the average
    amplitude
    for the 10 samples following the pick time

    :param tr: signal trace seismogram
    :type tr: obspy Trace
    :param pick_time: pick time
    :type pick_time: obspy UTCDateTime
    :param signal_type: Type of signal
    :type signal_type: str [accepted values are 'Acceleration', 'Velocity', and 'Displacement']
    :param nsamp_avg: number of sample on which to calculate the average difference
    :type nsamp_avg: int
    :returns: signa of signal polarity (1 or -1)
    :rtype: int
    """

    if signal_type not in ['Acceleration', 'Velocity', 'Displacement']:
        # raise error and exit funciton
        pass

    starttime = tr.stats.starttime
    sampling_rate = tr.stats.sampling_rate
    sample_pick = int(pick_time - starttime * sampling_rate)

    polarity = np.sign(np.mean(tr.data[sample_pick + 1:sample_pick +
                                                       nsamp_avg + 2] -
                               tr.data[sample_pick]))

    return polarity


def pick_uncertainty(tr, pick_time, snr_window=10):
    """Estimate the pick uncertainty, dt, based on the following equation
    dt = 1 / (fm * log_2(1 + SNR ** 2)
    where fm is the middle frequency and SNR, the signal to noise ratio in dB.

    :param tr: signal trace seismogram
    :type tr: obspy Trace
    :param pick_time: pick time
    :type pick_time: obspy UTCDateTime
    :param snr_window: length of window in ms centered on pick_time to
    calculate SNR.
        Noise and signal energy are calculated over the first and second half
        of the pick window, respectively.
    :type snr_window: float
    :returns: pick uncertainty in second
    :rtype: float
    """

    fm = 100  # central frequency in Hz
    # fm = centralfrequency(tr, pick_time, window)
    # tr is obspy tr, pick_time same, window window from pick over wich
    # signal frequencies are calculated
    st = Stream(traces=[tr])
    snr = calculate_snr(st, pick_time, snr_window)
    return 1 / (fm * np.log(1 + snr ** 2) / np.log(2))


def snr_repicker(st, picks, start_search_window, end_search_window,
                 search_resolution, pre_pick_window_len=1e-3,
                 post_pick_window_len=20e-3, trace_padding=30e-3):
    """
    Function to improve the picks based on the SNR.
    :param st: seismogram containing a seismic event
    :type st: :py:class:`obspy.core.stream.Stream`
    :param picks: list of uquake.core.event.Pick object
    picks
    :type picks: uquake.core.event.Catalog
    :param start_search_window: start of the search window relative to
    provided picks in seconds
    :type start_search_window: float
    :param end_search_window: end of the search window relative to the
    provided picks in seconds
    :type end_search_window: float
    :param search_resolution: resolution of the search space in seconds
    :type search_resolution: float
    :param pre_pick_window_len: length of the window before the presumed pick
    :type pre_pick_window_len: float
    :param post_pick_window_len: length of the window after the presumed pick
    :type post_pick_window_len: float
    :param trace_padding: buffer from trace start and end time excluded from
    search window
    :returns:  Tuple comprising 1) a :py:class:`uquake.core.event.Catalog`
    a new catalog containing a single event with a list of picks and 2) the SNR
    """

    snr_dt = np.arange(start_search_window, end_search_window,
                       search_resolution)

    st.detrend('demean')

    o_picks = []
    snrs = []

    pre_window_length = pre_pick_window_len
    post_window_length = post_pick_window_len

    for pick in picks:
        network = pick.waveform_id.network_code
        station = pick.waveform_id.station_code
        location = pick.waveform_id.location_code

        tr = st.select(network=network, station=station,
                       location=location).copy()[0]
        if tr is None:
            continue

        earliest_time = pick.time + start_search_window
        latest_time = pick.time + end_search_window

        if earliest_time is None or earliest_time < tr.stats.starttime + \
                trace_padding:
            earliest_time = tr.stats.starttime + trace_padding

        if latest_time is None or latest_time > tr.stats.endtime - \
                trace_padding:
            latest_time = tr.stats.endtime - trace_padding

        tau = []
        for dt in snr_dt:
            taut = pick.time + dt

            if earliest_time <= taut <= latest_time:
                tau.append(taut)

        if not tau:
            continue
        if tau[0] <= tr.stats.starttime:
            logger.error("Too early! tau[0]=%s <= tr.st=%s"
                         % (tau[0], tr.stats.starttime))
            logger.error("earliest_time:%s latest_time:%s"
                         % (earliest_time, latest_time))

            continue
        if tau[-1] >= tr.stats.endtime:
            logger.error("Too late! tau[-1]=%s >= tr.et=%s"
                         % (tau[-1], tr.stats.endtime))
            continue

        tau = np.array(tau)
        indices = (tau - tr.stats.starttime) * tr.stats.sampling_rate
        tmp = np.array([(taut, index,
                         calculate_snr(tr, taut,
                                       pre_wl=pre_window_length,
                                       post_wl=post_window_length))
                        for taut, index in zip(tau, indices)])

        alpha = 0

        index = np.argmax(tmp[:, 2])
        pick_time = tmp[index, 0]

        snr = calculate_snr(tr, pick_time, pre_wl=pre_window_length,
                            post_wl=post_window_length)

        # logger.debug("%s: sta:%s [%s] time_diff:%0.3f SNR:%.2f" %
        #              (function_name, station, phase, pick.time -
        #               pick_time, snr))

        method_string = 'snr_picker preWl=%.3g postWl=%.3g alpha=%.1f' % \
                        (pre_window_length, post_window_length, alpha)
        o_picks.append(make_pick(pick_time, phase=pick.phase_hint,
                                 wave_data=tr, snr=snr,
                                 method_string=method_string,
                                 resource_id=pick.resource_id))
        snrs.append(snr)

    return snrs, o_picks


def linearity_ensemble_re_picker(st, picks, start_search_window,
                                 end_search_window,
                                 start_refined_search_window,
                                 end_refinde_search_window,
                                 refined_window_search_resolution,
                                 linearity_calc_window_len):
    """
        Function to improve the picks based on the linearity or planarity
        for an ensemble of traces.
        :param st: waveforms containing a seismic event
        :type st: :py:class:`uquake.core.stream.Stream`
        :param picks: list of :class:`uquake.core.event.Pick` object
        picks
        :type picks: list of :py:class:`uquake.core.event.Pick`
        :param start_search_window: start of the search window relative to
        provided picks in seconds
        :type start_search_window: float
        :param end_search_window: end of the search window relative to the
        provided picks in seconds
        :param start_refined_search_window: start of the refined search
        window in seconds
        :type: float
        :param end_refined_search_window: end of the refined search window in
        seconds
        :type: float
        :type end_search_window: float
        :param refined_window_search_resolution: resolution of the search
        space in seconds
        :type refined_window_search_resolution: float
        :param linearity_calc_window_len: length of the window after the
        presumed pick used to calculate the linearity
        :param trace_padding: buffer from trace start and end time excluded
        from search window
        :returns:  Tuple comprising 1) a :py:class:`uquake.core.event.Catalog`
        a new catalog containing a single event with a list of picks and 2)
        the SNR
        """



def snr_ensemble_re_picker(st, picks, start_search_window, end_search_window,
                           start_refined_search_window,
                           end_refined_search_window,
                           refined_window_search_resolution,
                           snr_calc_pre_pick_window_len=1e-3,
                           snr_calc_post_pick_window_len=20e-3,
                           trace_padding=30e-3):
    """
    Function to improve the picks based on the SNR for an ensemble of traces.
    :param st: waveforms containing a seismic event
    :type st: :py:class:`uquake.core.stream.Stream`
    :param picks: list of :class:`uquake.core.event.Pick` object
    picks
    :type picks: list of :py:class:`uquake.core.event.Pick`
    :param start_search_window: start of the search window relative to
    provided picks in seconds
    :type start_search_window: float
    :param end_search_window: end of the search window relative to the
    provided picks in seconds
    :param start_refined_search_window: start of the refined search window in
    seconds
    :type: float
    :param end_refined_search_window: end of the refined search window in
    seconds
    :type: float
    :type end_search_window: float
    :param refined_window_search_resolution: resolution of the search space in
    seconds
    :type refined_window_search_resolution: float
    :param snr_calc_pre_pick_window_len: length of the window before the
    presumed pick
    :type snr_calc_pre_pick_window_len: float
    :param snr_calc_post_pick_window_len: length of the window after the
    presumed pick
    :type snr_calc_post_pick_window_len: float
    :param trace_padding: buffer from trace start and end time excluded from
    search window
    :returns:  Tuple comprising 1) a :py:class:`uquake.core.event.Catalog`
    a new catalog containing a single event with a list of picks and 2) the SNR
    """

    snr_dt = np.arange(start_refined_search_window, end_refined_search_window,
                       refined_window_search_resolution)

    refined_window_size = end_refined_search_window - \
                          start_refined_search_window

    biases = np.arange(start_search_window, end_search_window,
                       refined_window_size)

    st.detrend('demean')

    pre_window_length = snr_calc_pre_pick_window_len
    post_window_length = snr_calc_post_pick_window_len

    snrs_ensemble = []
    picks_ensemble = []
    for bias in biases:
        o_picks = []
        snrs = []
        for pick in picks:
            network = pick.waveform_id.network_code
            station = pick.waveform_id.station_code
            location = pick.waveform_id.location_code

            st_tmp = st.select(network=network, station=station,
                               location=location).copy()

            if len(st_tmp) == 0:
                continue

            tr = st_tmp.composite()[0]

            if tr is None:
                continue

            phase = pick.phase_hint

            earliest_time = pick.time + bias + start_refined_search_window
            latest_time = pick.time + bias + end_refined_search_window

            if earliest_time is None or earliest_time < tr.stats.starttime + \
                    trace_padding:
                earliest_time = tr.stats.starttime + trace_padding

            if latest_time is None or latest_time > tr.stats.endtime - \
                    trace_padding:
                latest_time = tr.stats.endtime - trace_padding

            tau = []
            for dt in (snr_dt + bias):
                taut = pick.time + dt

                if earliest_time <= taut <= latest_time:
                    tau.append(taut)

            if not tau:
                continue
            if tau[0] <= tr.stats.starttime:
                logger.error("Too early! tau[0]=%s <= tr.st=%s"
                             % (tau[0], tr.stats.starttime))
                logger.error("earliest_time:%s latest_time:%s"
                             % (earliest_time, latest_time))

                continue
            if tau[-1] >= tr.stats.endtime:
                logger.error("Too late! tau[-1]=%s >= tr.et=%s"
                             % (tau[-1], tr.stats.endtime))
                continue

            tau = np.array(tau)
            indices = (tau - tr.stats.starttime) * tr.stats.sampling_rate
            tmp = np.array([(taut, index,
                             calculate_snr(tr, taut,
                                           pre_wl=pre_window_length,
                                           post_wl=post_window_length))
                            for taut, index in zip(tau, indices)])

            alpha = 0

            index = np.argmax(tmp[:, 2])
            pick_time = tmp[index, 0]

            snr = calculate_snr(tr, pick_time, pre_wl=pre_window_length,
                                post_wl=post_window_length)

            method_string = 'snr_picker preWl=%.3g postWl=%.3g alpha=%.1f' % \
                            (pre_window_length, post_window_length, alpha)
            o_picks.append(make_pick(pick_time, phase=pick.phase_hint,
                                     wave_data=tr, snr=snr,
                                     method_string=method_string,
                                     resource_id=pick.resource_id))
            snrs.append(snr)

        snrs_ensemble.append(np.sum(snrs))
        picks_ensemble.append(o_picks)

    index = np.argmax(snrs_ensemble)
    output_picks = picks_ensemble[index]

    return snrs, output_picks


def extract_trace_segment(st: Stream, pick_time: UTCDateTime,
                          window_length: float):
    """
        Extract a fragment of trace and return the waveform in a matrix
        :param st: waveforms recorded by a particular site
        :type st: :py:class:`uquake.core.stream.Stream`
        :param pick_time: pick time
        :type pick_time: :py:class:uquake.core.UTCDateTime
        :param window_length: length of the window in seconds over which the
        linearity is measured
        :return: the waveform data package as a numpy ndarray
        :rtype: :py:class:numpy.ndarray
        """
    # Assuming here that all the traces recorded at a specific sensor are
    # aligned
    st[0].stats.sampling_rate

    waveforms = []
    for tr in st:
        sampling_rate = tr.stats.sampling_rate
        window_length_sample = window_length * sampling_rate
        start_measurement_window = (pick_time - tr.stats.sampling_rate)
        start_measurement_window_sample = start_measurement_window * \
                                          sampling_rate

        end_measurement_window_sample = start_measurement_window_sample + \
                                        window_length_sample
        waveforms.append(tr.data[start_measurement_window_sample:
                                 end_measurement_window_sample])

    return np.array(waveforms)



def measure_linearity(st: Stream, pick_time: UTCDateTime,
                      window_length: float):
    """
    measure the linearity of the particule motion using PCA
    :param st: waveforms recorded by a particular site
    :type st: :py:class:`uquake.core.stream.Stream`
    :param pick_time: pick time
    :type pick_time: :py:class:uquake.core.UTCDateTime
    :param window_length: length of the window in seconds over which the
    linearity is measured
    :return: the linearity between 0 and 1. 1 representing a perfect linearity
    :rtype: float
    """




    pass


def measure_planarity():
    """
    measure the planarity of the particule motion using PCA
    :param st: waveforms recorded by a particular site
    :type st: :py:class:`uquake.core.stream.Stream`
    :param pick_time: pick time
    :type pick_time: :py:class:uquake.core.UTCDateTime
    :param window_length: length of the window in seconds over which the
    linearity is measured
    :return: the linearity between 0 and 1. 1 representing a perfect planarity
    :rtype: float
    """
    pass


def calculate_snr(trace, pick, pre_wl=1e-3, post_wl=10e-3):
    """
    input :
    trs - Obspy stream
    Pick Time - in Obspy UTCDateTime
    preWl - Length of pre-window in seconds
    postWl - Length of post-window in seconds

    output:
    SNR - Signal to noise ratio
    """

    tr = trace

    sr = tr.stats.sampling_rate
    st = tr.stats.starttime
    et = tr.stats.endtime
    ps = int((pick - st) * sr)
    n_pre = int(pre_wl * sr)
    n_post = int(post_wl * sr)

    if pick + post_wl > et:
        energy_s = np.var(tr.data[ps:])
    else:
        energy_s = np.var(tr.data[ps:ps + n_post])

    if pick - pre_wl < st:
        energy_n = np.var(tr.data[:ps])
    else:
        energy_n = np.var(tr.data[ps - n_pre:ps])

    if (energy_n == 0) | (energy_s == 0):
        return 0

    snr = 10 * np.log10(energy_s / energy_n)

    return snr


def calculate_energy(stream, pick, Wl=5e-3):
    sr = stream.traces[0].stats['sampling_rate']
    st = stream.traces[0].stats['starttime']
    Ps = int((pick - st) * sr)
    Nb = int(Wl * sr)

    EnergyS = np.sum(
        [np.var(tr.data[Ps - Nb / 4:Ps + 3 * Nb / 2]) for tr in stream])

    return EnergyS


def _CalculateCF1_3(tr1, BW=None, WS=1e-3):
    if not BW:
        BW = [100, 5000]
    tr = tr1.copy()
    tr.taper(max_percentage=0.5, type='cosine')
    tr.filter(type='bandpass', freqmin=BW[0], freqmax=BW[1])

    cf1 = kurtosis(tr, win=WS)
    cf1 /= np.max(np.abs(cf1))

    cf2 = np.zeros(cf1.shape)
    dcf1 = np.diff(cf1)
    dcf1[dcf1 < 0] = 0
    cf2[0] = cf1[0]

    for k in range(1, len(cf1)):
        cf2[k] = cf2[k - 1] + dcf1[k - 1]

    try:
        cf3 = detrend(cf2, type='linear')
    except:
        cf3 = np.zeros(cf2.shape)

    return cf1, cf2, cf3


def measure_incidence_angle(st: Stream, inventory: Inventory,
                            picks: list, window_length_second: float):

    station_location_list = list(set([(tr.stats.station, tr.stats.location)
                                      for tr in st]))

    st_rotated = st.rotate('->ZNE', inventory=inventory)

    for station, location in station_location_list:
        st_ = st_rotated.select(station=station, location=location)
        p_pick = None
        s_pick = None
        for pick in picks:
            if (pick.waveform_id.station_code == station) and \
               (pick.waveform_id.location_code == location):

                if pick.phase_hint == 'P':
                    p_pick = pick

                elif pick.phase_hint == 'S':
                    s_pick = pick

        if p_pick is not None:
            st_.trim(starttime=p_pick.time,
                     endtime=p_pick.time + window_length_second)

            



